package com.example;

import akka.actor.typed.ActorRef;
import akka.actor.typed.Behavior;
import akka.actor.typed.javadsl.AbstractBehavior;
import akka.actor.typed.javadsl.ActorContext;
import akka.actor.typed.javadsl.Behaviors;
import akka.actor.typed.javadsl.Receive;

import java.util.Objects;

public class Actor extends AbstractBehavior<Actor.Message> {
    //TODO: Wartezeit bei der Operation
    //Zurückgegebener String von der linken Seite
    String leftSide;

    //Zurückgegebener String von der rechten Seite
    String rightSide;

    //Zeichen für die Operation (Add, Sub, Mul) welche benutzt wird
    String operationString;

    String name;
    Expression expression;
    public interface Message{}

    public record StartMessage(Expression expression) implements Message{}

    //Antwort von dem linken Kind
    public record LeftResponse(String string) implements Message{}

    //Antwort von dem rechten Kind
    public record RightResponse(String string) implements Message{}

    private Actor(ActorContext<Message> context, String name){
        super(context);
        this.name = name;
    }

    public static Behavior<Actor.Message> create(String name){
        return Behaviors.setup(context -> new Actor(context, name));
    }

    @Override
    public Receive<Actor.Message> createReceive(){
        return newReceiveBuilder()
                .onMessage(StartMessage.class, this::onStartMessage)
                .onMessage(LeftResponse.class, this::onLeftResponse)
                .onMessage(RightResponse.class, this::onRightResponse)
                .build();
    }

    private Behavior<Message> onStartMessage(StartMessage message){
        this.expression = message.expression;
        //Wenn initial nur Val übergeben wird → einfach ausgeben ansonsten UnterActors erstellen
        if (Objects.requireNonNull(expression) instanceof Expression.Add add) {
            this.operationString = "+";
            ActorRef<SubActor.Message> LeftSubActor = getContext().spawnAnonymous(SubActor.create());
            LeftSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),add.left(), true));
            ActorRef<SubActor.Message> RightSubActor = getContext().spawnAnonymous(SubActor.create());
            RightSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),add.right(), false));
        }
        else if (expression instanceof Expression.Sub sub) {
            this.operationString = "-";
            ActorRef<SubActor.Message> LeftSubActor = getContext().spawnAnonymous(SubActor.create());
            LeftSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),sub.left(), true));
            ActorRef<SubActor.Message> RightSubActor = getContext().spawnAnonymous(SubActor.create());
            RightSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),sub.right(), false));
        }
        else if (expression instanceof Expression.Mul mul) {
            this.operationString = "*";
            ActorRef<SubActor.Message> LeftSubActor = getContext().spawnAnonymous(SubActor.create());
            LeftSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),mul.left(), true));
            ActorRef<SubActor.Message> RightSubActor = getContext().spawnAnonymous(SubActor.create());
            RightSubActor.tell(new SubActor.PrintAndEvaluate(getContext().getSelf(),mul.right(), false));
        } else if (expression instanceof Expression.Val val) {
            System.out.println(val.inner());
        }
        return this;
    }

    private Behavior<Message> onLeftResponse(LeftResponse response){
        this.leftSide = response.string;
        if(this.rightSide != null){
            System.out.println("(" + leftSide + operationString + rightSide + ")");
        }
        return this;
    }

    private Behavior<Message> onRightResponse(RightResponse response){
        this.rightSide = response.string;
        if(this.leftSide != null){
            System.out.println("(" + leftSide + operationString + rightSide + ")");
        }
        return this;
    }
}
